(***************************************************************)
(* Coq code for OCAP                                           *)
(*                                                             *)
(* Typed Assembly Language                                     *)
(*                                                             *)
(* tal-0, a monomorphic version                                *)
(***************************************************************)

(* $Id: tal0.v,v 1.4 2006/07/15 02:56:41 ssg05 Exp $ *)

Require Import FMapB.
Require Import MAP.
Require Import machine.
Require Import natutil2.
Require Import extac.

Open Local Scope tm_scope.

Inductive flag : Set :=
  | init : flag
  | junk : flag.

Inductive valTy : Set :=
  | valty_var : nat -> valTy
  | valty_int : valTy
  | valty_tup : tupTy -> valTy
  | valty_code : rfTy -> valTy
with rfTy : Set := 
  | rfty_nil : rfTy
  | rfty_cons : (prod reg valTy) -> rfTy -> rfTy
with tupTy : Set :=
  | tupty_nil : tupTy
  | tupty_cons : (prod flag valTy) -> tupTy 
                  -> tupTy.

Implicit Types Y Gamma : rfTy.
Implicit Types t tau : valTy.
Implicit Types o tup : tupTy.

Scheme valTy_ind4 :=
  Induction for valTy Sort Prop
 with rfTy_ind4 :=
  Induction for rfTy Sort Prop
 with tupTy_ind4 :=
  Induction for tupTy Sort Prop.

(* **************************************************)
(* cell type                                        *)

Definition cellTy := prod flag valTy.

(* **************************************************)
(* code type                                        *)

Definition codeTy := rfTy.


(* **************************************************)
(* Notations                                        *)

Notation " ^ n " := (valty_var n) 
  (at level 40) : tal_scope.
Open Local Scope tal_scope.

Notation " 'int' " := (valty_int)
  (at level 0) : tal_scope.

Notation " []- Y " := (valty_code Y)
  (at level 40) : tal_scope.

Notation "[{ r , .. , r' }]" := 
  (rfty_cons r .. (rfty_cons r' rfty_nil) ..) 
  (at level 0) : tal_scope.

Notation "[{}]" := (rfty_nil)  
  (at level 0) : tal_scope.

Notation "[< o , .. , o' >]" := 
  (valty_tup (tupty_cons o .. (tupty_cons o' tupty_nil) ..)) 
  (at level 0) : tal_scope.

(* **************************************************)
(* regfile type                                     *)

Module RfTy.

Fixpoint sel (m : rfTy) (r : reg) {struct m} : option valTy :=
  match reg.beq_reg r zero with
    | true => Some valty_int
    | false =>
      match m with 
        | rfty_nil => None
        | rfty_cons (r', b') m' =>
          match reg.beq_reg r r' with
            | true => Some b'
            | false => sel m' r
          end
      end
  end.

Fixpoint upd (m : rfTy) (r : reg) (b : valTy) {struct m} : rfTy :=
  match reg.beq_reg r zero with
    | true => m
    | false =>
      match m return rfTy with
        | rfty_nil => rfty_cons (r, b) rfty_nil
        | rfty_cons (r', b') m' =>
          match reg.beq_reg r r' with
            | true => rfty_cons (r, b) m'
            | false => rfty_cons (r', b') (upd m' r b)
          end
      end
  end.

Definition indom (m : rfTy) (r : reg) := 
  exists b : valTy, sel m r = Some b.

Definition lookup (m : rfTy) (r : reg) (t : valTy) := 
  sel m r = Some t.

Lemma lookup_nil :
  forall r t,
    beq_reg r zero = false
    ->lookup rfty_nil r t 
    -> False.
Proof.
  intros r t Hr H.
  unfold lookup in H.
  unfold sel in H.
  rewrite Hr in H.
  discriminate.
Qed.

Lemma indom_nil :
  forall r,
    beq_reg r zero = false
    -> indom rfty_nil r -> False.
Proof.
  unfold indom; simpl; intros r Hr [b H].
  rewrite Hr in H; discriminate.
Qed.

Lemma sel_upd_req :
  forall Y r r' t,
    beq_reg zero r = false
    -> beq_reg r r' = true
    -> sel (upd Y r' t) r = Some t.
Proof.
  intros Y r r' t H H1.
  induction Y as [|(r0, t0) Y IHY].
  unfold upd, sel.
    breg_simpl.
    trivial.
  unfold upd, sel.
    breg_simpl.
    destruct (reg.beq_dec r' r0) as [Heq | Hneq].
    breg_simpl.
    trivial.
    breg_simpl.
    fold sel; fold upd.
    apply IHY.
Qed.

Lemma sel_upd_rneq :
  forall Y r r' t,
    beq_reg r r' = false
    -> sel (upd Y r' t) r = sel Y r.
Proof.
  intros Y r r' t H.
  induction Y as [|(rx, tx) Y IHY].
  simpl.
    destruct (reg.beq_dec r r0) as [He | He]; 
      rewrite He; trivial.
    assert (H0 : beq_reg r' r0 = false); 
      [breg_solve | rewrite H0; clear H0].
    simpl.
   breg_simpl; trivial.
   destruct (reg.beq_dec r' r0) as [He1 | He1]; 
     rewrite He1; trivial; simpl; breg_simpl; trivial.
  simpl.
    destruct (reg.beq_dec r rx) as [He | He];
      rewrite He; trivial.
    assert (H0 : beq_reg r' rx = false); 
      [breg_solve | rewrite H0; clear H0].
    destruct (reg.beq_dec r r0) as [He' | He'];
      rewrite He'; trivial.
    assert (H0 : beq_reg r' r0 = false); 
      [breg_solve | rewrite H0; clear H0].
    simpl; breg_simpl; trivial.
    destruct (reg.beq_dec r' r0) as [He1 | He1]; 
      rewrite He1; trivial; simpl; breg_simpl; trivial.
    destruct (reg.beq_dec r' r0) as [He1 | He1]; 
      rewrite He1; trivial; simpl; breg_simpl; trivial.
    destruct (reg.beq_dec r r0) as [He2 | He2]; 
      rewrite He2; trivial; simpl; breg_simpl; trivial.
    destruct (reg.beq_dec r' rx) as [He3 | He3]; 
      rewrite He3; trivial; simpl; breg_simpl; trivial.
    destruct (reg.beq_dec r' rx) as [He3 | He3]; 
      rewrite He3; trivial; simpl; breg_simpl; trivial.
Qed.

End RfTy.


(* **************************************************)
(* tuple type                                       *)

Module TupTy.

Fixpoint mem (tup : tupTy) (ct : cellTy) {struct tup} : Prop := 
  match tup with
    | tupty_nil => False
    | tupty_cons ct' tup' => ct = ct' \/ mem tup' ct
  end.

Fixpoint lookup (tup : tupTy) (n:nat) (ct:cellTy) 
  {struct n} : Prop :=
  match n, tup with
    | O, tupty_cons x l' => x = ct
    | O, tupty_nil => False
    | S m, tupty_nil => False
    | S m, tupty_cons x l' => lookup l' m ct
  end.

Fixpoint get (tup : tupTy) (n : nat) 
  {struct n} : option cellTy :=
  match n, tup with
    | O, tupty_cons ct l' => Some ct
    | O, tupty_nil => None
    | S m, tupty_nil => None
    | S m, tupty_cons x l' => get l' m
  end.
  
Fixpoint upd (tup : tupTy) (n : nat) (ct : cellTy) 
  {struct n}: tupTy :=
  match n, tup with 
    | O, tupty_cons ct' l' => tupty_cons ct l'
    | S m, tupty_cons ct' l' => tupty_cons ct' (upd l' m ct)
    | _, _ => tupty_nil
  end.

Definition indom (tup : tupTy) (n : nat) := 
  exists ct : cellTy, lookup tup n ct.

Lemma lookup_nil : 
  forall n ct,
    lookup tupty_nil n ct -> False.
Proof.
  intros n ct.
  intros Hf.
  induction n; simpl; intros; auto.
Qed.

Lemma lookup_get : forall n l ct,
  lookup l n ct -> get l n = Some ct.
Proof.
  induction n; destruct l; 
    simpl; intros;
      tauto || auto.
  subst; trivial.
Qed.

Lemma get_lookup : forall n l ct,
  get l n = Some ct -> lookup l n ct.
Proof.
  induction n; destruct l; 
    simpl; intros;
      discriminate || auto.
  injection H; intro; trivial.
Qed.

Lemma lookup_upd_indom :
  forall n l n' (ct ct' : cellTy),
    lookup (upd l n ct) n' ct' -> indom l n'.
Proof.
  unfold indom;
  induction n; destruct l; destruct n'; 
    simpl; intros;
    tauto || auto.
  exists p; trivial.
  exists ct'; trivial.
  exists p; trivial.
  eapply IHn; eauto.
Qed.

Lemma upd_a_lookup_a : 
  forall n l (ct ct' : cellTy),
    lookup (upd l n ct') n ct -> ct = ct'.
Proof.
  induction n; destruct l; simpl; intros;
    try tauto.
  subst ct; trivial.
  eapply IHn; eauto.
Qed.

Lemma upd_a_lookup_a' : 
  forall addr addr' l (ct ct' : cellTy),
    addr <> addr'
    -> lookup (upd l addr ct') addr' ct 
       = lookup l addr' ct.
Proof.
  induction addr; destruct l; simpl in * ;
    try tauto.
  induction addr'; intros.
  destruct (H (refl_equal _)).
  simpl.
  trivial.
  intros.
  induction addr'; intros.
  simpl. trivial.
  simpl.
  apply IHaddr.
  intro.
  apply H.
  rewrite H0.
  trivial.
Qed.

Lemma lookup_unique :
  forall addr l  (ct ct' : cellTy),
    lookup l addr ct 
    -> lookup l addr ct'
    -> ct = ct'.
Proof.
  induction addr; destruct l; simpl; intros .
  destruct H.
  subst p; trivial.
  destruct H.
  apply (IHaddr l ct ct' H H0).
Qed.

Lemma lookup_upd : 
  forall addr l (ct ct' : cellTy),
    lookup l addr ct
    -> lookup (upd l addr ct') addr ct'.
Proof.
  induction addr; destruct l; simpl; intros; trivial.
  eapply IHaddr; eauto.
Qed.

Lemma lookup_upd_neq :
  forall addr l addr' (ct ct' : cellTy),
    addr <> addr'
    -> lookup l addr ct
    -> lookup (upd l addr' ct') addr ct.
Proof.
  induction addr; destruct l; simpl; intros; 
    tauto || trivial.
  subst p.
  destruct addr'.
  tauto.
  simpl in * .
  trivial.
  destruct addr'.
  simpl.
  assumption.
  simpl.
  apply IHaddr.
  intro Hf.
  apply H.
  subst addr; trivial.
  trivial.
Qed.

Lemma indom_indom_upd :
  forall n (tup : tupTy) (n': nat) (ct : cellTy),
    indom tup n -> indom (upd tup n' ct) n.
Proof.
  unfold indom.
  intros n tup n' ct [ct' H].
  destruct (eq_dec n n') as [Heq | Hneq].
  subst n'.
  exists ct.
  eapply lookup_upd; eauto.
  exists ct'.
  assert (H1 : n' <> n).
  intro H0; apply (Hneq (sym_eq H0)).
  apply lookup_upd_neq; trivial.
Qed.

Lemma indom_upd_indom :
  forall n (tup : tupTy) (n': nat) (ct : cellTy),
    indom (upd tup n' ct) n -> indom tup n.
Proof.
  intros n tup n' ct [ct' H].
  eapply lookup_upd_indom; eauto.
Qed.

End TupTy.

(* ************************************************ *)
(* Psi : CHSpec                                     *)
(* Phi : DHSPec                                     *)


Module CdSpec <: ONE_SET.
  Definition B := codeTy.
End CdSpec.
Module CHSpecDef := FuncMap CdSpec.
Module CHSpec := MapEx CHSpecDef.
(* psi : chspec *)
Definition chspec := CHSpec.map.

Implicit Types Psi psi: chspec.

Module DtSpec <: ONE_SET.
  Definition B := cellTy.
End DtSpec.
Module DHSpecDef := FuncMap DtSpec.
Module DHSpec := MapEx DHSpecDef.
(* psi : chspec *)
Definition dhspec := DHSpec.map.

Implicit Types Phi phi: dhspec.

(* **************************************************)
(* cell subtyping                                   *)

Inductive ceTy_sub : forall (ct ct' : cellTy), Prop :=
  | ceTy_sub_refl : forall (ct : cellTy),
                       ceTy_sub ct ct
  | ceTy_sub_init : forall (t : valTy), 
                       ceTy_sub (init, t) (junk, t).

Lemma ceTy_sub_trans : forall ct ct' ct'',
  ceTy_sub ct ct'
  -> ceTy_sub ct' ct''
  -> ceTy_sub ct ct''.
Proof.
  intros ct ct' ct'' H12 H23.
  destruct ct; destruct ct'; destruct ct''.
  destruct f1; inversion_clear H12;  inversion_clear H23;
    try apply ceTy_sub_refl ; try apply ceTy_sub_init.
Qed.

Lemma ceTy_sub_init_ct : forall t ct ct',
  ceTy_sub ct' ct
  -> ceTy_sub (init, t) ct
  -> ceTy_sub (init, t) ct'.
Proof.
  intros; destruct ct'; destruct ct.
  inversion_clear H; inversion_clear H0;
    try apply ceTy_sub_init;
      try apply ceTy_sub_refl.
Qed.

(* **************************************************)
(* regfile subtyping                                *)

(*  TODO: very weak subtyping rule *)
Inductive rfTy_sub : forall (Y Y' : rfTy), Prop :=
  | rfty_sub :
      forall (Y Y' : rfTy),
        (forall r : reg,
             RfTy.indom Y' r
             -> RfTy.sel Y' r = RfTy.sel Y r)
        -> rfTy_sub Y Y'.

Lemma rfty_sub_refl : forall Y, rfTy_sub Y Y.
Proof.
  intros.
  apply rfty_sub; trivial.
Qed.

Lemma rfty_sub_trans : forall Y Y' Y'',
  rfTy_sub Y Y'
  -> rfTy_sub Y' Y''
  -> rfTy_sub Y Y''.
Proof.
  intros Y Y' Y'' H12 H23.
  inversion_clear H12.
  inversion_clear H23.
  apply rfty_sub; trivial.
  intros r Hindom.
  inversion Hindom.
  rewrite (H0 r Hindom) in H1.
  assert (H3 : RfTy.indom Y' r).
  unfold RfTy.indom.
  exists x.
  apply H1.
  rewrite (H0 r Hindom).
  apply (H r H3).
Qed.

Lemma rfty_lookup_sub_lookup : forall Y Y' r w,
  RfTy.lookup Y' r w -> rfTy_sub Y Y' 
  -> RfTy.lookup Y r w.
Proof.
   intros Y Y' r w HY' Hsub.
   inversion_clear Hsub.
   unfold RfTy.indom, RfTy.lookup in * .
   assert (H0 : exists b, RfTy.sel Y' r = Some b).
   exists w; trivial.
   rewrite (H r H0) in HY'.
   assumption.
Qed.

(* **************************************************)
(* code subtyping                                   *)

(*  |- a1 <= a2 *)

Inductive codeTy_sub : forall (a a' : codeTy), Prop :=
  | codety_sub : 
        forall  (Y Y' : rfTy),
          rfTy_sub Y Y'
          -> codeTy_sub Y Y'.


(* **************************************************)
(* value typing                                     *)

Inductive valok : chspec -> dhspec -> word -> valTy -> Prop :=
  | valok_lab : 
              forall (Psi : chspec) (Phi : dhspec) 
                (f : word) (Y : rfTy),
                CHSpec.lookup Psi f Y
                -> valok Psi Phi f (valty_code Y)
  | valok_int : forall (Psi : chspec) (Phi : dhspec) (w : word),
              valok Psi Phi w valty_int
  | valok_tup : forall (Psi : chspec) (Phi : dhspec) 
                    (l : label) (tup : tupTy),
              (forall i, 
                TupTy.indom tup i
                -> exists ct, exists ct',   
                  TupTy.lookup tup i ct
                  /\ DHSpec.lookup Phi (l+i) ct'
                  /\ ceTy_sub ct' ct)
              -> valok Psi Phi l (valty_tup tup).
              
Inductive regok : rfTy -> reg -> valTy -> Prop :=
  | regok_reg : forall (Y : rfTy) (r : reg) (t : valTy),
              RfTy.sel Y r = Some t
              -> regok Y r t.

Inductive ceok : chspec -> dhspec -> word -> cellTy -> Prop :=
  | ceok_junk : forall Psi Phi v (t : valTy),
                  ceok Psi Phi v (junk, t)
  | ceok_init : forall Psi Phi v t,
                  valok Psi Phi v t
                   -> ceok Psi Phi v (init, t).


(* **************************************************)
(* instruction typing                               *)

Inductive wfis : forall (Psi : chspec)
  (Gamma :rfTy) (I : instrseq), Prop :=

  | wfis_j : 
          forall Psi Gamma Gamma'' f,
            CHSpec.lookup Psi f Gamma''
            -> codeTy_sub Gamma Gamma''
            -> wfis Psi Gamma (ione (j f))

  | wfis_jr : 
          forall Psi Gamma Gamma'' r,
            regok Gamma r (valty_code Gamma'')
            -> codeTy_sub Gamma Gamma''
            -> wfis Psi Gamma (ione (jr r))

  | wfis_addu : 
          forall Psi Gamma rd rs rt I',
            beq_reg rd zero = false
            -> beq_reg rs zero = false
            -> regok Gamma rs valty_int
            -> regok Gamma rt valty_int
            -> wfis Psi (RfTy.upd Gamma rd valty_int) I'
            -> wfis Psi Gamma (iseq (addu rd rs rt) I')

  | wfis_addu_mov : 
          forall Psi Gamma t rd rs rt I',
            beq_reg rd zero = false
            -> beq_reg rs zero = true
            -> regok Gamma rt t
            -> wfis Psi (RfTy.upd Gamma rd t) I'
            -> wfis Psi Gamma (iseq (addu rd rs rt) I')

  | wfis_addiu : 
          forall Psi Gamma rd rs w I',
            beq_reg rd zero = false
            -> beq_reg rs zero = false
            -> regok Gamma rs valty_int
            -> wfis Psi (RfTy.upd Gamma rd valty_int) I'
            -> wfis Psi Gamma (iseq (addiu rd rs w) I')

  | wfis_addiu_movi : 
          forall Psi Gamma Gamma'' rd rs f I',
            beq_reg rd zero = false
            -> beq_reg rs zero = true
            -> CHSpec.lookup Psi f Gamma''
            -> wfis Psi (RfTy.upd Gamma rd (valty_code Gamma'')) I'
            -> wfis Psi Gamma (iseq (addiu rd rs f) I')

  | wfis_subu : 
          forall Psi Gamma rd rs rt I',
            beq_reg rd zero = false
            -> regok Gamma rs valty_int
            -> regok Gamma rt valty_int
            -> wfis Psi (RfTy.upd Gamma rd valty_int) I'
            -> wfis Psi Gamma (iseq (subu rd rs rt) I')

  | wfis_subiu : 
          forall Psi Gamma rd rs w I',
            beq_reg rd zero = false
            -> regok Gamma rs valty_int
            -> wfis Psi (RfTy.upd Gamma rd valty_int) I'
            -> wfis Psi Gamma (iseq (subiu rd rs w) I')

  | wfis_lw : 
          forall Psi Gamma o t rt w rs I',
            beq_reg rt zero = false
            -> regok Gamma rs (valty_tup o)
            -> TupTy.lookup o w (init, t)
            -> wfis Psi (RfTy.upd Gamma rt t) I'
            -> wfis Psi Gamma (iseq (lw rt w rs) I')

  | wfis_sw :
          forall Psi Gamma o ct t rt w rs I',
            beq_reg rs zero = false
            -> regok Gamma rs (valty_tup o)
            -> TupTy.lookup o w ct 
            -> ceTy_sub (init, t) ct
            -> regok Gamma rt t
            -> wfis Psi 
                (RfTy.upd Gamma rs 
                (valty_tup (TupTy.upd o w (init, t)))) I'
            -> wfis Psi Gamma (iseq (sw rt w rs) I')

  | wfis_blt :
          forall Psi Gamma Gamma'' rs rt f I',
            CHSpec.lookup Psi f Gamma''
            -> regok Gamma rs valty_int
            -> regok Gamma rt valty_int
            -> codeTy_sub Gamma Gamma''
            -> wfis Psi Gamma I'
            -> wfis Psi Gamma (iseq (blt rs rt f) I')

  | wfis_bgtz :
          forall Psi Gamma Gamma'' rs f I',
            CHSpec.lookup Psi f Gamma''
            -> regok Gamma rs valty_int
            -> codeTy_sub Gamma Gamma''
            -> wfis Psi Gamma I'
            -> wfis Psi Gamma (iseq (bgtz rs f) I').


(* **************************************************)
(* codeheap typing                                  *)

Definition WF_C (Psi Psi' : chspec) (C : codeheap) : Prop :=
  forall (f : label) (Gamma : rfTy),
    CHSpec.lookup Psi' f Gamma
    -> (exists I : _ , 
        codeheap.getis C f I
        /\ wfis Psi Gamma I).

(* **************************************************)
(* regfile typing                                   *)

Definition WF_R Psi Phi Gamma R : Prop :=
  forall r t, 
    RfTy.sel Gamma r = Some t
    -> valok Psi Phi (Regfile.get R r) t.

Definition WF_H Psi Phi H : Prop :=
  forall l ct,
    DHSpec.lookup Phi l ct
    -> exists v,
        Heap.lookup H l v
        /\ ceok Psi Phi v ct.

Definition WF_S Psi Gamma (S : state) : Prop :=
  let (H, R, pc) := S in
    exists Phi : dhspec,
      WF_H Psi Phi H /\ WF_R Psi Phi Gamma R.

(* **************************************************)
(* program typing                                   *)

Definition WF_P (Psi : chspec) (Gamma : rfTy) (P : machine) : Prop :=
  let (C, S) := P in
    let (H, R, pc) := S in
      WF_C Psi Psi C
      /\ (exists Phi : dhspec,
        WF_H Psi Phi H /\ WF_R Psi Phi Gamma R)
      /\ (exists I : instrseq, 
        (codeheap.getis C pc I) 
        /\ wfis Psi Gamma I).

Set Implicit Arguments.
Unset Strict Implicit.

Lemma reg_typing : 
  forall Psi Phi Gamma R r t,
    WF_R Psi Phi Gamma R
    -> regok Gamma r t
    -> valok Psi Phi (Regfile.get R r) t.
Proof.
  intros Psi Phi Gamma R r t HR Hr.
  destruct Hr as [Gamma r t H].
  unfold WF_R in HR.
  auto.
Qed.

Lemma regfile_update :
  forall Psi Phi Gamma R r w t,
    reg.beq_reg r zero = false
    -> WF_R Psi Phi Gamma R
    -> valok Psi Phi w t
    -> WF_R Psi Phi (RfTy.upd Gamma r t) (Regfile.set R r w).
Proof.
  intros Psi Phi Gamma R r w t Hzero HR Hw.
  unfold WF_R in * .
  intros r0 t0.
  assert (HR' := HR r0 t0).
  destruct (reg.beq_dec r0 r) as [Heq | Hneq].
  breg_simpl.
    rewrite (RfTy.sel_upd_req Gamma r r t 
      Hzero (beq_refl r)).
    intros Ht; injection Ht; intro; subst t0.
    regfile.simplgs.
    trivial.
  regfile.simplgs.
    intro Ht.
    rewrite (RfTy.sel_upd_rneq Gamma r0 r t Hneq) in Ht.
    apply (HR' Ht).
Qed.
Implicit Arguments regfile_update
  [Psi Phi Gamma R r w t].

(* Lemma WF_R_weaken_Gamma :  *)
Lemma regfile_weakening :
  forall Psi Phi Gamma Gamma' R,
    WF_R Psi Phi Gamma R
    -> rfTy_sub Gamma Gamma'
    -> WF_R Psi Phi Gamma' R.
Proof.
  intros Psi Phi Gamma Gamma' R HR Hsub.
  unfold WF_R in * .
  intros r t Hr.
  inversion_clear Hsub as [Y Y' Hin].
  unfold RfTy.indom in Hin.
  apply HR.
  apply sym_eq.
  assert (Hin' := Hin r).
  rewrite Hr in Hin'.
  apply Hin'.
  exists t.
  trivial.
Qed.
Implicit Arguments regfile_weakening 
   [Psi Phi Gamma Gamma' R].

Lemma heapval_exists:
  forall Psi Phi H o ct addr offset,
    WF_H Psi Phi H
    -> TupTy.lookup o offset ct
    -> valok Psi Phi addr (valty_tup o)
    -> exists v, 
      Heap.lookup H (addr+offset) v.
Proof.
  intros Psi Phi H o ct0 addr offset.
  intros HH Ht Ho.
  inversion_clear Ho.
  assert (Hin : TupTy.indom o offset).
  unfold TupTy.indom.
  exists ct0; trivial.
  destruct (H0 offset Hin) as [ct [ct' [Hct [ Hct' Hsub]]]].
  assert (He := TupTy.lookup_unique _ _ _ _ Hct Ht).
  subst ct0.
  unfold WF_H in HH.
  clear Hin.
  destruct (HH (addr+offset) ct' Hct') as
    [v [Hin Hceok]].
  exists v.
  trivial.
Qed.
Implicit Arguments heapval_exists
    [Psi Phi H o ct addr offset].

(* Lemma heap_val_typing :  *)
Lemma heapval_typing :
  forall Psi Phi H o t addr offset,
    WF_H Psi Phi H
    -> TupTy.lookup o offset (init, t)
    -> valok Psi Phi addr (valty_tup o)
    -> exists v, Heap.lookup H (addr+offset) v
      /\ valok Psi Phi v t.
Proof.
  intros Psi Phi H o t addr offset.
  intros HH Ht Ho.
  inversion_clear Ho.
  assert (Hin : TupTy.indom o offset).
  unfold TupTy.indom.
  exists (init, t); trivial.
  destruct (H0 offset Hin) as [ct [ct' [Hct [ Hct' Hsub]]]].
  assert (He := TupTy.lookup_unique _ _ _ _ Hct Ht).
  subst ct.
  unfold WF_H in HH.
  clear Hin.
  destruct (HH (addr+offset) ct' Hct') as
    [v [Hin Hceok]].
  exists v.
  split_l.
  inversion Hsub.
  subst ct'; subst ct.
  inversion_clear Hceok.
  trivial.
Qed.
Implicit Arguments heapval_typing 
    [ Psi Phi H o t addr offset].

Lemma heap_typing_reg_tying_ex_t :
  forall Psi Phi Gamma H R r o offset ct,
    WF_H Psi Phi H
    -> WF_R Psi Phi Gamma R
    -> regok Gamma r (valty_tup o)
    -> TupTy.lookup o offset ct
    -> exists ct',
      DHSpec.lookup Phi (Regfile.get R r + offset) ct'
      /\ ceTy_sub ct' ct.
Proof.
  intros Psi Phi Gamma H R r o offset ct.
  intros HH HR Hr Hct.
  inversion_clear Hr.
  unfold WF_R in HR.
  substH HR with (HR r (valty_tup o) H0).
  inversion_clear HR.
  assert (Hin : TupTy.indom o offset).
  unfold TupTy.indom.
  exists ct; trivial.
  substH H1 with (H1 offset Hin).
  destruct H1 as [ctx [ctx' [Hctx [Hctx' Hsub]]]].
  exists ctx'.
  split_l.
  rewrite (TupTy.lookup_unique _ _ _ _ Hct Hctx).
  trivial.
Qed.
Implicit Arguments heap_typing_reg_tying_ex_t
    [Psi Phi Gamma H R r o offset ct].

Lemma heap_extend_valok : 
  forall Psi Phi addr ct t v,
    DHSpec.get Phi addr = None
    -> valok Psi Phi v t
    -> valok Psi (DHSpec.set Phi addr ct) v t.
Proof.
  intros Psi Phi addr ct t v.
  intros Hnin Hv.
  inversion_clear Hv.
  apply valok_lab; trivial.
  apply valok_int; trivial.
  apply valok_tup.
  intros i Hindom.
  substH H with (H i Hindom).
  destruct H as [ctx [ctx' [Hctx [Hctx' Hsub]]]].
  destruct (eq_dec addr (v + i)) as [Heq | Hneq].
(* case addr = v + i *)
  subst addr.
  rewrite Hctx' in Hnin.
  discriminate Hnin.
(* case addr <> v + i *)
  exists ctx. exists ctx'.
  split_l.
  split.
  apply DHSpec.get_lookup.
  rewrite (DHSpec.set_a_get_a' Phi 
    ct Hneq).
  apply (DHSpec.lookup_get Hctx').
  assumption.
Qed.
Implicit Arguments heap_extend_valok
  [Psi Phi addr ct t v].

Lemma heap_update_valok : 
  forall Psi Phi Phi' addr ct ct' t v,
    DHSpec.lookup Phi addr ct
    -> ceTy_sub ct' ct
    -> DHSpec.canupd Phi addr ct' Phi'
    -> valok Psi Phi v t
    -> valok Psi Phi' v t.
Proof.
  intros Psi Phi Phi' addr ct ct' t v.
  intros Hct Hsub Hphi' Hvalok.
  inversion_clear Hvalok.
  apply valok_lab; trivial.
  apply valok_int; trivial.
  apply valok_tup.
  intros i Hindom.
  substH H with (H i Hindom).
  destruct H as [ctx [ctx' [Hctx [Hctx' Hsubx]]]].
  destruct (eq_dec addr (v + i)) as [Heq | Hneq].
(* case addr = v + i *)
  subst addr.
  exists ctx. exists ct'.
  split_l.
  split.
  apply (DHSpec.upd_lookup Hphi').
  assert (Hg := DHSpec.upd_a_get_a Hphi').
  substH Hct with (DHSpec.get_lookup
    (m:=Phi) Hct).
  assert (Heq := DHSpec.lookup_unique Hctx' Hct).
  subst ctx'.
  apply (ceTy_sub_trans ct' ct ctx Hsub Hsubx).
(* case addr <> v + i *)
  assert (Hg := DHSpec.upd_a_get_a' Hneq Hphi').
  exists ctx. exists ctx'.
  split_l.
  split.
  apply DHSpec.get_lookup.
  rewrite Hg.
  apply (DHSpec.lookup_get Hctx').
  assumption.
Qed.
Implicit Arguments heap_update_valok
  [Psi Phi Phi' addr ct ct' t v].

Lemma heap_extend_ceok :
  forall Psi Phi addr ct ct' v,
    DHSpec.get Phi addr = None
    -> ceok Psi Phi v ct
    -> ceok Psi (DHSpec.set Phi addr ct') v ct.
Proof.
  intros Psi Phi addr ct ct' v.
  intros Hnin Hceok.
  inversion_clear Hceok.
  apply ceok_junk; trivial.
  apply ceok_init; trivial.
  destruct ct'.
  eapply heap_extend_valok; eauto.
Qed.
Implicit Arguments heap_extend_ceok
  [Psi Phi addr ct ct' v].
  
Lemma heap_update_ceok :
  forall Psi Phi Phi' addr ct ct' ct'' v,
    DHSpec.lookup Phi addr ct
    -> ceTy_sub ct' ct
    -> DHSpec.canupd Phi addr ct' Phi'
    -> ceok Psi Phi v ct''
    -> ceok Psi Phi' v ct''.
Proof.
  intros Psi Phi Phi' addr ct ct' ct0 v.
  intros Hct Hsub Hphi' Hceok.
  inversion_clear Hceok.
  apply ceok_junk; trivial.
  apply ceok_init; trivial.
  eapply heap_update_valok; eauto.
Qed.
Implicit Arguments heap_update_ceok
  [Psi Phi Phi' addr ct ct' ct'' v].
  
Lemma heap_extend_WF_H :
  forall Psi Phi H ct addr w,
    WF_H Psi Phi H
    -> DHSpec.get Phi addr = None
    -> ceok Psi (DHSpec.set Phi addr ct) w ct
    -> WF_H Psi (DHSpec.set Phi addr ct)
            (Heap.set H addr w).
Proof.
  intros Psi Phi H ct addr w.
  intros HH Hnin Hw.
  unfold WF_H in * .
  intros l ct0 Hct0.
  destruct (eq_dec addr l) as [Heq | Hneq].
(* case l = addr *)
  subst addr.
  exists w.
  substH Hct0 with (DHSpec.lookup_get Hct0).
  rewrite (DHSpec.set_a_get_a Phi l ct) in Hct0.
  inj_hyp Hct0.
  subst ct0.
  split_r.
  apply Heap.get_lookup.
  apply Heap.set_a_get_a.
(* case l <> addr *)
  substH Hct0 with (DHSpec.lookup_get Hct0).
  rewrite (DHSpec.set_a_get_a' Phi ct Hneq) in Hct0.
  substH Hct0 with (DHSpec.get_lookup Hct0).
  destruct (HH l ct0 Hct0) as [v [Hv Hceok]].
  exists v.
  split.
  apply Heap.get_lookup.
  rewrite <- (Heap.lookup_get Hv).
  apply (Heap.set_a_get_a' H w Hneq).
  apply (heap_extend_ceok (ct':=ct) Hnin Hceok).
Qed.
Implicit Arguments heap_extend_WF_H
    [Psi Phi H ct addr w].

Lemma heap_update_WF_H :
  forall Psi Phi Phi' H ct ct' addr w,
    WF_H Psi Phi H
    -> ceTy_sub ct' ct
    -> DHSpec.lookup Phi addr ct
    -> DHSpec.canupd Phi addr ct' Phi'
    -> ceok Psi Phi' w ct'
    -> WF_H Psi Phi' (Heap.set H addr w).
Proof.
  intros Psi Phi Phi' H ct ct' addr w.
  intros HH Hsub Hct Hphi' Hw.
  unfold WF_H in * .
  intros l ct0 Hct0.
  destruct (eq_dec addr l) as [Heq | Hneq].
(* case l = addr *)
  subst l.
  assert (Hg := DHSpec.get_lookup (m := Phi) Hct).
  substH Hphi' with (DHSpec.upd_lookup Hphi').
  assert (Hcteq := DHSpec.lookup_unique Hct0 Hphi').
  subst ct0.
  exists w.
  split_r.
  apply Heap.get_lookup.
  apply Heap.set_a_get_a.
(* case l <> addr *)
  assert (H0 := DHSpec.upd_a_get_a' Hneq Hphi').
  substH Hct0 with (DHSpec.lookup_get (m := Phi') Hct0).
  rewrite Hct0 in H0; clear Hct0.
  substH HH with (HH l ct0 
    (DHSpec.get_lookup (m:=Phi) (sym_eq H0))).
  destruct HH as [v [Hv Hceok]].
  exists v.
  split.
  apply Heap.get_lookup.
  rewrite <- (Heap.lookup_get Hv).
  apply (Heap.set_a_get_a' H w Hneq).
  apply (heap_update_ceok Hct Hsub Hphi' Hceok).
Qed.
Implicit Arguments heap_update_WF_H
    [Psi Phi Phi' H ct ct' addr w].

Lemma heap_extend_WF_R :
  forall Psi Phi Gamma R ct addr,
    WF_R Psi Phi Gamma R
    -> DHSpec.get Phi addr = None
    -> WF_R Psi (DHSpec.set Phi addr ct) Gamma R.
Proof.
  intros Psi Phi Gamma R ct addr.
  intros HR Hnin.
  unfold WF_R in * .
  intros r t Hr.
  apply heap_extend_valok; trivial.
  apply HR; trivial.
Qed.
Implicit Arguments heap_extend_WF_R 
    [Psi Phi Gamma R ct addr].

Lemma heap_update_WF_R :
  forall Psi Phi Phi' Gamma R ct ct' addr,
    WF_R Psi Phi Gamma R
    -> DHSpec.lookup Phi addr ct
    -> DHSpec.canupd Phi addr ct' Phi'
    -> ceTy_sub ct' ct
    -> WF_R Psi Phi' Gamma R.
Proof.
  intros Psi Phi Phi' Gamma R ct ct' addr.
  intros HR Hct Hphi' Hsub.
  unfold WF_R in * .
  intros r t Hr.
  apply heap_update_valok with (Phi:=Phi) (addr:=addr) 
    (ct := ct) (ct' := ct'); trivial.
  apply HR; trivial.
Qed.
Implicit Arguments heap_update_WF_R 
    [Psi Phi Phi' Gamma R ct ct' addr].

(* Lemma heap_update_valok_strong :  *)
Lemma heap_update_tuple_valok :
  forall Psi Phi Phi' addr offset o ct t,
    valok Psi Phi addr (valty_tup o)
    -> TupTy.lookup o offset ct
    -> ceTy_sub (init, t) ct
    -> DHSpec.canupd Phi (addr + offset) (init, t) Phi'
    -> valok Psi Phi' addr 
           (valty_tup (TupTy.upd o offset (init, t))).
Proof.
  intros Psi Phi Phi' addr offset tup ct t.
  intros Hvalok Htup Hsub Hphi'.
  inversion_clear Hvalok.
  apply valok_tup.
  intros i Hindom. 
  assert (Hindom' : TupTy.indom tup i).
  eapply TupTy.indom_upd_indom; eauto.
  substH H with (H i Hindom').
  destruct H as [ctx [ctx' [Hctx [Hctx' Hsubx]]]].
  destruct (eq_dec i offset) as [Heq | Hneq].
  subst i.
  exists (init, t). exists (init, t).
  split_l using (apply (TupTy.lookup_upd _ _ _ (init, t) Hctx)).
  split_l using (apply (DHSpec.canupd_lookup Hphi')).
  apply ceTy_sub_refl.
  exists ctx. exists ctx'.
  split_l using 
    (apply (TupTy.lookup_upd_neq _ _ _ _ (init, t) Hneq Hctx)).
  split_r.
  assert (Hneq' : addr + i <> addr + offset).
  intro Hf; apply Hneq.
  eapply Arith.Plus.plus_reg_l; eauto.
  apply (DHSpec.canupd_lookup_neq Hneq' Hphi' Hctx').
Qed.
Implicit Arguments heap_update_tuple_valok 
   [Psi Phi Phi' addr offset o ct t].

Lemma heap_update_WF_R_reg :
  forall Psi Phi Phi' Gamma R o ct ct' t r offset,
    beq_reg r zero = false
    -> WF_R Psi Phi Gamma R
    -> regok Gamma r (valty_tup o)
    -> TupTy.lookup o offset ct
    -> DHSpec.lookup Phi 
         (Regfile.get R r + offset) ct'
    -> DHSpec.canupd Phi 
         (Regfile.get R r + offset) (init, t) Phi'
    -> ceTy_sub ct' ct
    -> ceTy_sub (init, t) ct'
    -> WF_R Psi Phi' (RfTy.upd Gamma r 
         (valty_tup (TupTy.upd o offset (init, t)))) R.
Proof.
  intros Psi Phi Phi' Gamma R o ct ct' t r offset.
  intros Hnr0 HR Hr Hct Hctin Hphi' Hsub Hsub'.
  unfold WF_R in * .
  intros r0 t0 Hr0.
  destruct (reg.beq_dec r0 r) as [Heq | Hneq].
(* case : r = r0 *)
  rewrite (RfTy.sel_upd_req Gamma r0 r 
    (valty_tup (TupTy.upd o offset (init, t))))
    in Hr0; [idtac|breg_solve|assumption].
  inj_hyp Hr0.
  subst t0.
  assert (H : valok Psi Phi (Regfile.get R r) 
      (valty_tup o)).
  inversion_clear Hr.
  apply (HR r (valty_tup o) H).
  assert (Hsub'' : ceTy_sub (init, t) ct).
  eapply ceTy_sub_trans; eauto.
  rewrite (beq_true_req r0 r Heq).
  apply (heap_update_tuple_valok H Hct Hsub'' Hphi').
(* case : r <> r0 *)
  rewrite (RfTy.sel_upd_rneq Gamma r0 r 
    (valty_tup (TupTy.upd o offset (init, t))) Hneq)
    in Hr0.
  apply (heap_update_valok Hctin Hsub' Hphi' 
    (HR r0 t0 Hr0)).
Qed.
Implicit Arguments heap_update_WF_R_reg
    [Psi Phi Phi' Gamma R o ct ct' t r offset].

Theorem Soundness :
  forall Psi (Gamma : rfTy) P,
    WF_P Psi Gamma P ->
    (exists P' : _, (P ~~> P')
      /\ (exists Gamma' : _,  WF_P Psi Gamma' P')).
Proof.
  intros Psi Gamma P HP.
  unfold WF_P in HP.
  destruct P as [C [H R pc]].
  destruct HP as [HC [[Phi [HH HR]] [I [Hpc HI]]]].
  destruct I.
  destruct i as
    [ rd rs rt | rd rs w 
      | rd rs rt | rd rs w
      | rt i rs | rt i rs
      | rs rt f | rs f
      | f | r]; 
    inversion_clear HI as 
      [ p0 g0 Gamma'' f0 Hf Hsub
        | p0 g0 Gamma'' r0 Hr Hsub
        | p0 g0 r0 r1 r2 i0 Hrdnr0 Hrsnr0 Hrs Hrt HI'
        | p0 g0 t r0 r1 r2 i0 Hrdnr0 Hrsr0 Hrt HI'
        | p0 g0 r0 r1 w0 i0 Hrdnr0 Hrsnr0 Hrs HI'
        | p0 g0 Gamma'' r0 r1 w0 i0 Hrdnr0 Hrsr0 Hf HI'
        | p0 g0 r0 r1 r2 i0 Hrdnr0 Hrs Hrt HI'
        | p0 g0 r0 r1 w0 i0 Hrdnr0 Hrs HI'
        | p0 g0 o t r0 w0 r1 i0 Hrtnr0 Hrs Ht HI'
        | p0 g0 o ct t r0 w0 r1 i0 Hrsnr0 Hrs Hct Hsub Hrt HI'
        | p0 g0 Gamma'' r0 r1 f0 i0 Hf Hrs Hrt Hsub HI'
        | p0 g0 Gamma'' r0 f0 i0 Hf Hrs Hsub HI'
      ].
  (* case jd f *)
  exists (C, <.H, R, f.>).
  inversion_clear Hpc.
  unfold step. 
  rewrite H0.
  simpl.
  split_l.
  inversion_clear Hsub.
  exists Gamma''.
  split_l.
  split.
  exists Phi.
  split_l.
  eapply regfile_weakening; eauto.
  apply (HC f Gamma'' Hf).
  (* case wfis_jr r *)
  exists (C, <.H, R, Regfile.get R r.>).
  inversion_clear Hpc.
  unfold step.
  rewrite H0.
  simpl.
  split_l.
  exists Gamma''.
  split_l.
  split.
  exists Phi.
  split_l.
  eapply regfile_weakening; eauto.
  inversion_clear Hsub; trivial.
  unfold WF_C in HC.
  inversion_clear Hr.
  unfold WF_R in HR.
  assert (H3 := HR r (valty_code Gamma'') H1).
  inversion_clear H3.
  apply (HC (Regfile.get R r) Gamma'' H2).
  (* case iseq *)
  destruct i as
    [ rd rs rt | rd rs w 
      | rd rs rt | rd rs w
      | rt i rs | rt i rs
      | rs rt f | rs f
      | f | r]; 
    inversion_clear HI as 
      [ p0 g0 Gamma'' f0 Hf Hsub
        | p0 g0 Gamma'' r0 Hr Hsub
        | p0 g0 r0 r1 r2 i0 Hrdnr0 Hrsnr0 Hrs Hrt HI'
        | p0 g0 t r0 r1 r2 i0 Hrdnr0 Hrsr0 Hrt HI'
        | p0 g0 r0 r1 w0 i0 Hrdnr0 Hrsnr0 Hrs HI'
        | p0 g0 Gamma'' r0 r1 w0 i0 Hrdnr0 Hrsr0 Hf HI'
        | p0 g0 r0 r1 r2 i0 Hrdnr0 Hrs Hrt HI'
        | p0 g0 r0 r1 w0 i0 Hrdnr0 Hrs HI'
        | p0 g0 o t r0 w0 r1 i0 Hrtnr0 Hrs Ht HI'
        | p0 g0 o ct t r0 w0 r1 i0 Hrsnr0 Hrs Hct Hsub Hrt HI'
        | p0 g0 Gamma'' r0 r1 f0 i0 Hf Hrs Hrt Hsub HI'
        | p0 g0 Gamma'' r0 f0 i0 Hf Hrs Hsub HI'
      ].
  (* case wfis_addu *)
  exists (C, <.H, Regfile.set R rd 
    (Regfile.get R rs + Regfile.get R rt), ++pc .>).
  inversion_clear Hpc.
  unfold step.
  rewrite H0.
  simpl.
  split_l.
  exists (RfTy.upd Gamma rd valty_int).
  split_l.
  split.
  exists Phi.
  split_l.
  eapply regfile_update; eauto; apply valok_int.
  exists I;  split_lr.
  (* case wfis_addu_mov *)
  exists (C, <.H, Regfile.set R rd 
    (Regfile.get R rs + Regfile.get R rt), ++pc .>).
  inversion_clear Hpc as [ | | Hx0 Hx1 Hx2 Hx3 Hi HCI].
  unfold step.
  rew_clear Hi.
  simpl.
  split_l.
  exists (RfTy.upd Gamma rd t).
  split_l.
  split.
  exists Phi.
  split_l.
  breg_simpl.
  regfile.simplgs.
  simpl.
  inversion_clear Hrt.
  eapply regfile_update; eauto.
  exists I;  split_lr.
  (* case wfis_addiu *)
  exists (C, <.H, Regfile.set R rd 
    (Regfile.get R rs + w), ++pc .>).
  inversion_clear Hpc as [ | | Hx0 Hx1 Hx2 Hx3 Hi HCI].
  unfold step.
  rew_clear Hi.
  simpl.
  split_l.
  exists (RfTy.upd Gamma rd valty_int).
  split_l.
  split.
  exists Phi.
  split_l.
  eapply regfile_update; eauto; 
    apply valok_int.
  exists I;  split_lr.
  (* case wfis_addiu_movi *)
  exists (C, <.H, Regfile.set R rd 
    (Regfile.get R rs + w), ++pc .>).
  inversion_clear Hpc as [ | | Hx0 Hx1 Hx2 Hx3 Hi HCI].
  unfold step.
  rew_clear Hi.
  simpl.
  split_l.
  exists (RfTy.upd Gamma rd (valty_code Gamma'')).
  split_l.
  split.
  exists Phi.
  split_l.
  breg_simpl.
  regfile.simplgs.
  simpl.
  eapply regfile_update; eauto; 
    apply valok_lab; trivial.
  exists I; split_lr.
  (* case wfis_subu *)
  exists (C, <.H, Regfile.set R rd 
    (Regfile.get R rs - Regfile.get R rt), ++pc .>).
  inversion_clear Hpc.
  unfold step.
  rewrite H0.
  simpl.
  split_l.
  exists (RfTy.upd Gamma rd valty_int).
  split_l.
  split.
  exists Phi.
  split_l.
  eapply regfile_update; eauto; apply valok_int.
  exists I;  split_lr.
  (* case wfis_subiu *)
  exists (C, <.H, Regfile.set R rd 
    (Regfile.get R rs - w), ++pc .>).
  inversion_clear Hpc as [ | | Hx0 Hx1 Hx2 Hx3 Hi HCI].
  unfold step.
  rew_clear Hi.
  simpl.
  split_l.
  exists (RfTy.upd Gamma rd valty_int).
  split_l.
  split.
  exists Phi.
  split_l.
  eapply regfile_update; eauto; apply valok_int.
  exists I;  split_lr.
  (* case wfis_lw *)
  destruct (heapval_typing HH Ht (reg_typing HR Hrs))
    as [v [Hin Hv]].
  exists (C, <.H, Regfile.set R rt v, ++pc .>).
  inversion_clear Hpc as [ | | Hx0 Hx1 Hx2 Hx3 Hi HCI].
  unfold step.
  rew_clear Hi.
  simpl.
  rewrite Hin.
  split_l.
  exists (RfTy.upd Gamma rt t).
  split_l.
  split.
  exists Phi.
  split_l.
  apply regfile_update; auto.
  exists I; split_lr.
  (* case wfis_sw *)
  set (H' := Heap.set H (Regfile.get R rs + i) 
     (Regfile.get R rt)).
  exists (C, <.H', R, ++pc .>).
  inversion_clear Hpc as [ | | Hx0 Hx1 Hx2 Hx3 Hi HCI].
  unfold step.
  rew_clear Hi.
  simpl.
  assert (Hval :=reg_typing HR Hrs).
  destruct (heapval_exists HH Hct Hval)
    as [v Hvin].
  rewrite Hvin.
  split_l.
  exists (RfTy.upd Gamma rs (valty_tup 
    (TupTy.upd o i (init, t)))).
  destruct (heap_typing_reg_tying_ex_t
    HH HR Hrs Hct) as [ct' [Hct'in Hsub']].
  set (addr := (Regfile.get R rs + i)) in * .
  set (Phi' := DHSpec.set Phi addr (init, t)).
  split_l.
  split.
    exists Phi'.
    assert (Hx := DHSpec.lookup_canupd (m:=Phi) (a:=addr) 
      (b:= ct') (init, t) Hct'in).
    assert (Hsubx := ceTy_sub_init_ct t ct ct' Hsub' Hsub).
    split.
    refine (heap_update_WF_H HH Hsubx Hct'in Hx _).
    apply ceok_init.
    eapply reg_typing; eauto.
    eapply heap_update_WF_R; eauto.
    eapply heap_update_WF_R_reg; eauto.
  exists I.
  split_lr.
  (* case wfis_blt *)
  inversion_clear Hpc as [ | | Hx0 Hx1 Hx2 Hx3 Hi HCI].
  unfold step, nextS.
  rew_clear Hi.
  bnat_blt_dec (Regfile.get R rs) (Regfile.get R rt).
  exists (C, <.H, R, f.>).
  split_l.
  inversion_clear Hsub.
  exists Gamma''.
  split_l.
  split.
    exists Phi.
    split_l.
    eapply regfile_weakening; eauto.
  apply (HC f Gamma'' Hf).

  exists (C, <.H, R, pc+1.>).
  split_l.
  exists Gamma.
  split_l.
  split.
    exists Phi.
    split_lr.
  exists I; split_lr.
  (* case wfis_bgtz *)
  inversion_clear Hpc as [ | | Hx0 Hx1 Hx2 Hx3 Hi HCI].
  unfold step, nextS.
  rew_clear Hi.
  bnat_blt_dec O (Regfile.get R rs).
  exists (C, <.H, R, f.>).
  split_l.
  inversion_clear Hsub.
  exists Gamma''.
  split_l.
  split.
    exists Phi.
    split_l.
    eapply regfile_weakening; eauto.
  apply (HC f Gamma'' Hf). 
  
  exists (C, <.H, R, pc+1.>).
  split_l.
  exists Gamma.
  split_l.
  split.
    exists Phi.
    split_lr.
  exists I; split_lr.
Qed.

(* a simple type checker *)
Fixpoint typechk (psi : chspec) (a : codeTy) (is : instrseq) 
  {struct is} : Prop :=
  match is with 
    | ione (j f) => 
        match (CHSpec.get psi f) with
          | Some a'' => codeTy_sub a a''
          | _ => False
        end

    | ione (jr r) => 
        match (RfTy.sel a r) with
          | Some (valty_code a'') => codeTy_sub a a''
          | _ => False
        end

    | iseq (addu rd rs rt) is' => 
        match (beq_reg rd zero, beq_reg rs zero) with
          | (false, false) => 
            match (RfTy.sel a rs, RfTy.sel a rt) with
              | (Some valty_int, Some valty_int) => 
                   typechk psi (RfTy.upd a rd valty_int) is'
              | _ => False
            end
          | (false, true) => 
            match RfTy.sel a rt with 
              | Some t => 
                  typechk psi (RfTy.upd a rd t) is'
              | _ => False
            end
          | _ => False 
        end

    | iseq (addiu rd rs w) is' => 
        match (beq_reg rd zero, beq_reg rs zero) with
          | (false, false) => 
            match RfTy.sel a rs with
              | Some valty_int => 
                   typechk psi (RfTy.upd a rd valty_int) is'
              | _ => False
            end
          | (false, true) => 
            match (CHSpec.get psi w) with 
              | Some a'' => 
                  typechk psi (RfTy.upd a rd (valty_code a'')) is'
              | _ => False
            end
          | _ => False 
        end

    | iseq (subu rd rs rt) is' => 
        match beq_reg rd zero with
          | false => 
            match (RfTy.sel a rs, RfTy.sel a rt) with
              | (Some valty_int, Some valty_int) => 
                   typechk psi (RfTy.upd a rd valty_int) is'
              | _ => False
            end
          | _ => False 
        end

    | iseq (subiu rd rs w) is' => 
        match beq_reg rd zero with
          | false => 
            match RfTy.sel a rs with
              | Some valty_int => 
                   typechk psi (RfTy.upd a rd valty_int) is'
              | _ => False
            end
          | _ => False 
        end

    | iseq (lw rt w rs) is' => 
        match beq_reg rt zero with
          | false => 
            match RfTy.sel a rs with
              | Some (valty_tup o) =>
                match (TupTy.get o w) with 
                  | Some (init, t) => 
                    typechk psi (RfTy.upd a rt t) is'
                  | _ => False
                end
              | _ => False
            end
          | _ => False 
        end

    | iseq (sw rt w rs) is' => 
        match beq_reg rs zero with
          | false => 
            match (RfTy.sel a rs, RfTy.sel a rt) with
              | (Some (valty_tup o), Some t) =>
                match (TupTy.get o w) with 
                  | Some (init, t') => 
                    t' = t 
                    /\ typechk psi (RfTy.upd a rs 
                      (valty_tup (TupTy.upd o w (init, t)))) is'
                  | Some (junk, t') => 
                    t' = t 
                    /\ typechk psi (RfTy.upd a rs 
                      (valty_tup (TupTy.upd o w (init, t)))) is'
                  | _ => False
                end
              | _ => False
            end
          | _ => False 
        end

    | iseq (blt rs rt f) is' => 
        match CHSpec.get psi f with
          | Some a'' => 
            match (RfTy.sel a rs, RfTy.sel a rt) with
              | (Some valty_int, Some valty_int) => 
                  codeTy_sub a a''
                  /\ typechk psi a is'
              | _ => False
            end
          | _ => False 
        end

    | iseq (bgtz rs f) is' => 
        match CHSpec.get psi f with
          | Some a'' => 
            match RfTy.sel a rs with
              | Some valty_int => 
                  codeTy_sub a a''
                  /\ typechk psi a is'
              | _ => False
            end
          | _ => False 
        end

    | _ => False
  end.

Theorem type_checker_sound : forall psi is a,
  typechk psi a is -> wfis psi a is.
Proof.
  intros psi is.
  induction is.
  destruct i as
    [ rd rs rt | rd rs w 
      | rd rs rt | rd rs w
      | rt i rs | rt i rs
      | rs rt f | rs f
      | f | r]; 
    simpl; intros a H; 
      try contradiction.
  
  (* case i = j f *)
  assert (Hf : exists a'', 
    CHSpec.get psi f = Some a'');
    [destruct (CHSpec.get psi f) as [a'' |] ;
      try contradiction;
        exists a''; trivial
      | destruct Hf as [a'' Hf];
        rewrite Hf in H].
  eapply wfis_j; eauto.
  apply CHSpec.get_lookup; trivial.

  (* case i = jr r *)
  assert (Hr : exists t, 
    RfTy.sel a r = Some t);
    [destruct (RfTy.sel a r) as [t |] ;
      try contradiction;
        exists t; trivial
      | destruct Hr as [t Hr];
        rewrite Hr in H].
  destruct t as [ | | o | a'' ]; try contradiction.
  apply wfis_jr with a''; trivial;
    try constructor; trivial.

  (* case is = iseq i is' *)
  destruct i as
    [ rd rs rt | rd rs w 
      | rd rs rt | rd rs w
      | rt i rs | rt i rs
      | rs rt f | rs f
      | f | r]; 
    simpl; intros a H; try contradiction.

  (* case i = addu rd rs rt *)
  assert (Hrdnr0 : beq_reg rd zero = false);
    [destruct (beq_reg rd zero); 
      [contradiction | trivial]
      | rewrite Hrdnr0 in H].
  destruct (reg.beq_dec rs zero) as [Hrsr0 |Hrsr0];
    rewrite Hrsr0 in H.

  assert (Hrt : exists t, 
    RfTy.sel a rt = Some t);
    [destruct (RfTy.sel a rt) as [t |] ;
      try contradiction;
        exists t; trivial
      | destruct Hrt as [t Hrt];
        rewrite Hrt in H].
  apply wfis_addu_mov with t; try constructor; trivial.
  eapply IHis; eauto.

  assert (Hrs : exists t, 
    RfTy.sel a rs = Some t);
    [destruct (RfTy.sel a rs) as [t |] ;
      try contradiction;
        exists t; trivial
      | destruct Hrs as [t Hrs];
        rewrite Hrs in H].
  destruct t as [ | | o | a'' ]; try contradiction.
  assert (Hrt : exists t, 
    RfTy.sel a rt = Some t);
    [destruct (RfTy.sel a rt) as [t |] ;
      try contradiction;
        exists t; trivial
      | destruct Hrt as [t Hrt];
        rewrite Hrt in H].
  destruct t as [ | | o | a'' ]; try contradiction.
  apply wfis_addu; try constructor; trivial.
  eapply IHis; eauto.

  (* case i = addiu rd rs w *)
  assert (Hrdnr0 : beq_reg rd zero = false);
    [destruct (beq_reg rd zero); 
      [contradiction | trivial]
      | rewrite Hrdnr0 in H].
  destruct (reg.beq_dec rs zero) as [Hrsr0 |Hrsr0];
    rewrite Hrsr0 in H.

  assert (Hw : exists a'', 
    CHSpec.get psi w = Some a'');
    [destruct (CHSpec.get psi w) as [a'' |] ;
      try contradiction;
        exists a''; trivial
      | destruct Hw as [a'' Hw];
        rewrite Hw in H].
  apply wfis_addiu_movi with a''; 
    try constructor; trivial.
  eapply IHis; eauto.

  assert (Hrs : exists t, 
    RfTy.sel a rs = Some t);
    [destruct (RfTy.sel a rs) as [t |] ;
      try contradiction;
        exists t; trivial
      | destruct Hrs as [t Hrs];
        rewrite Hrs in H].
  destruct t as [ | | o | a'' ]; try contradiction.
  apply wfis_addiu; try constructor; trivial.
  eapply IHis; eauto.
  
  (* case i = subu rd rs rt *)
  assert (Hrdnr0 : beq_reg rd zero = false);
    [destruct (beq_reg rd zero); 
      [contradiction | trivial]
      | rewrite Hrdnr0 in H].
  assert (Hrs : exists t, 
    RfTy.sel a rs = Some t);
    [destruct (RfTy.sel a rs) as [t |] ;
      try contradiction;
        exists t; trivial
      | destruct Hrs as [t Hrs];
        rewrite Hrs in H].
  destruct t as [ | | o | a'' ]; try contradiction.
  assert (Hrt : exists t, 
    RfTy.sel a rt = Some t);
    [destruct (RfTy.sel a rt) as [t |] ;
      try contradiction;
        exists t; trivial
      | destruct Hrt as [t Hrt];
        rewrite Hrt in H].
  destruct t as [ | | o | a'' ]; try contradiction.
  apply wfis_subu; try constructor; trivial.
  eapply IHis; eauto.

  (* case i = subiu rd rs w *)
  assert (Hrdnr0 : beq_reg rd zero = false);
    [destruct (beq_reg rd zero); 
      [contradiction | trivial]
      | rewrite Hrdnr0 in H].
  assert (Hrs : exists t, 
    RfTy.sel a rs = Some t);
    [destruct (RfTy.sel a rs) as [t |] ;
      try contradiction;
        exists t; trivial
      | destruct Hrs as [t Hrs];
        rewrite Hrs in H].
  destruct t as [ | | o | a'' ]; try contradiction.
  apply wfis_subiu; try constructor; trivial.
  eapply IHis; eauto.

  (* case i = lw rt w rs *)
  assert (Hrtnr0 : beq_reg rt zero = false);
    [destruct (beq_reg rt zero); 
      [contradiction | trivial]
      | rewrite Hrtnr0 in H].
  assert (Hrs : exists t, 
    RfTy.sel a rs = Some t);
    [destruct (RfTy.sel a rs) as [t |] ;
      try contradiction;
        exists t; trivial
      | destruct Hrs as [t Hrs];
        rewrite Hrs in H].
  destruct t as [ | | o | a'' ]; try contradiction.
  assert (Hw : exists ct, 
    TupTy.get o i = Some ct);
    [destruct (TupTy.get o i) as [ct |] ;
      try contradiction;
        exists ct; trivial
      | destruct Hw as [ct Hw];
        rewrite Hw in H].
  destruct ct as [f t]; destruct f.
  apply wfis_lw with o t; 
    try constructor; trivial.
  apply TupTy.get_lookup; trivial.
  eapply IHis; eauto.
  contradiction.

  (* case i = sw rt w rs *)
  assert (Hrsnr0 : beq_reg rs zero = false);
    [destruct (beq_reg rs zero); 
      [contradiction | trivial]
      | rewrite Hrsnr0 in H].
  assert (Hrs : exists t, 
    RfTy.sel a rs = Some t);
    [destruct (RfTy.sel a rs) as [t |] ;
      try contradiction;
        exists t; trivial
      | destruct Hrs as [t Hrs];
        rewrite Hrs in H].
  destruct t as [ | | o | ax ]; try contradiction.
  assert (Hrt : exists t, 
    RfTy.sel a rt = Some t);
    [destruct (RfTy.sel a rt) as [t |] ;
      try contradiction;
        exists t; trivial
      | destruct Hrt as [t Hrt];
        rewrite Hrt in H].
  assert (Hw : exists ct, 
    TupTy.get o i = Some ct);
    [destruct (TupTy.get o i) as [ct |] ;
    try contradiction;
      exists ct; trivial
      | destruct Hw as [ct Hw];
        rewrite Hw in H].
  destruct ct as [f t']; destruct f.

  destruct H as [Heq H].
  subst t'.
  apply wfis_sw with o (init, t) t; 
    try constructor; trivial.
  apply TupTy.get_lookup; trivial.
  eapply IHis; eauto.

  destruct H as [Heq H].
  subst t'.
  apply wfis_sw with o (junk, t) t; 
    try constructor; trivial.
  apply TupTy.get_lookup; trivial.
  eapply IHis; eauto.

  (* case i = blt rs rt f *)
  assert (Hf : exists a'', 
    CHSpec.get psi f = Some a'');
    [destruct (CHSpec.get psi f) as [a'' |] ;
      try contradiction;
        exists a''; trivial
      | destruct Hf as [a'' Hf];
        rewrite Hf in H].
  assert (Hrs : exists t, 
    RfTy.sel a rs = Some t);
    [destruct (RfTy.sel a rs) as [t |] ;
      try contradiction;
        exists t; trivial
      | destruct Hrs as [t Hrs];
        rewrite Hrs in H].
  destruct t as [ | | o | ax ]; try contradiction.
  assert (Hrt : exists t, 
    RfTy.sel a rt = Some t);
    [destruct (RfTy.sel a rt) as [t |] ;
      try contradiction;
        exists t; trivial
      | destruct Hrt as [t Hrt];
        rewrite Hrt in H].
  destruct t as [ | | o | ax ]; try contradiction.
  destruct H as [Hsub H].
  apply wfis_blt with a''; trivial;
    try constructor; trivial.
  eapply IHis; eauto.

  (* case i = bgtz rs rt f *)
  assert (Hf : exists a'', 
    CHSpec.get psi f = Some a'');
    [destruct (CHSpec.get psi f) as [a'' |] ;
      try contradiction;
        exists a''; trivial
      | destruct Hf as [a'' Hf];
        rewrite Hf in H].
  assert (Hrs : exists t, 
    RfTy.sel a rs = Some t);
    [destruct (RfTy.sel a rs) as [t |] ;
      try contradiction;
        exists t; trivial
      | destruct Hrs as [t Hrs];
        rewrite Hrs in H].
  destruct t as [ | | o | ax ]; try contradiction.
  destruct H as [Hsub H].
  apply wfis_bgtz with a''; trivial;
    try constructor; trivial.
  eapply IHis; eauto.

Qed.

Ltac typecheck :=
  match goal with 
    | |- wfis ?psi ?a ?is =>
      apply type_checker_sound;
        compute;
          simpl;
            match goal with
              | |- False => fail 1 "Ill-formed type."
              | _ => idtac
            end
    | _ => fail 1 "The goal is not a TAL instrseq judgement."
  end.

Unset Implicit Arguments.
