(***************************************************************)
(* Coq code for OCAP                                           *)
(*                                                             *)
(* MAP Interface                                               *)
(*                                                             *)
(***************************************************************)

(* $Id: MAP.v,v 1.4 2006/07/15 02:56:42 ssg05 Exp $ *)

Notation Local optionX := option.

Notation Local SomeX := Some.

Notation Local NoneX := None.

Notation Local B_U := Set.

Require Import extac.

Set Implicit Arguments.
Unset Strict Implicit.

Module Type ONE_SET.

  Parameter B : Set.

End ONE_SET.

Module Type MAPDEF.

  Parameter A : Set.

  Parameter beq : A -> A -> bool.

  Axiom beq_true_eq : forall a a' : A, 
    beq a a' = true -> a = a'.

  Axiom beq_false_neq : forall a a' : A,
    beq a a' = false -> a <> a'.

  Axiom eq_beq_true : forall a a' : A, 
    a = a' -> beq a a' = true.

  Axiom neq_beq_false : forall a a' : A, 
    a <> a' -> beq a a' = false.
  
  Axiom eq_dec : forall a a' : A, {a=a'}+{a<>a'}.

  Parameter B : B_U.

(***********************)
(** map                *)
(***********************)

  Parameter map : B_U.
  
  Parameter emp_map : map.

  Parameter sig_map : A -> B -> map.

(***********************)
(** get upd set        *)
(***********************)

  Parameter get : map -> A -> optionX B.

  Axiom get_dec : forall (m : map) (a : A), 
    {exists b, get m a = SomeX b} + {get m a = NoneX}.

  Parameter upd : map -> A -> B -> optionX map.

  Parameter set : map -> A -> B -> map.

  Axiom set_sem : forall m a a' b, 
    match (beq a a') with
      | true => get (set m a b) a = SomeX b
      | false => get (set m a b) a' = get m a'
    end.

  Axiom upd_sem : forall m a b,
    match get_dec m a with 
      | left _ => upd m a b = SomeX (set m a b)
      | right _ => upd m a b = NoneX
    end.

  Axiom emp_sem : forall a : A, get emp_map a = NoneX.
  
  Axiom sig_sem : forall (a a' : A) (b : B), 
    get (sig_map a' b) a = 
    match (beq a a') with
      | true => SomeX b
      | false => NoneX
    end.


(************************)
(** merge               *)
(************************)

  Parameter merge : map -> map -> map.

  Axiom merge_sem : forall (m1 m2 : map) (a : A),
    get (merge m1 m2) a 
    = match get m1 a, get m2 a with
        | SomeX b1, SomeX b2 => SomeX b1
        | SomeX b1, NoneX => SomeX b1
        | NoneX, SomeX b2 => SomeX b2
        | NoneX, NoneX => NoneX
      end.

End MAPDEF.

Module MapEx (MapDef : MAPDEF).

Import MapDef.

Load MAPINC.

Definition upd : map -> A -> B -> optionX map
  := MapDef.upd.

Lemma upd_sem : forall m a b,
  match get_dec m a with 
    | left _ => upd m a b = SomeX (set m a b)
    | right _ => upd m a b = NoneX
  end.
Proof MapDef.upd_sem.

Definition canupd (m : map) (a : A) (b : B) (m' : map) :=
  upd m a b = SomeX m'.

Lemma upd_a_get_a : forall (m m' : map) (a : A) (b : B),
  upd m a b = SomeX m'
  -> exists b', get m a = SomeX b'.
Proof.
  intros m m' a b Hu.
  assert (H := upd_sem m a b).
  destruct (get_dec m a).
  assumption.
  rewrite H in Hu; discriminate.
Qed.

Lemma get_a_upd_a : forall (m : map) (a : A) (b b' : B),
  get m a = SomeX b
  -> upd m a b' = SomeX (set m a b').
Proof.
  intros.
  assert (H0 := upd_sem m a b').
  destruct (get_dec m a).
  assumption.
  rewrite e in H.
  discriminate.
Qed.

Lemma upd_a_get_a' : forall (m m' : map) (a a' : A) (b : B),
  a <> a'
  -> upd m a b = SomeX m'
  -> get m' a' = get m a'.
Proof.
  intros m m' a a' b H H0.
  assert (H1 := upd_sem m a b).
  destruct (get_dec m a).
  rewrite H0 in H1.
  injection H1; intro He; subst m'.
  apply set_a_get_a'; trivial.
  rewrite H0 in H1.
  discriminate.
Qed.

Lemma canupd_get : forall (m m' : map) (a : A) (b : B),
  canupd m a b m'
  -> exists b', get m a = SomeX b'.
Proof.
  unfold canupd; apply upd_a_get_a.
Qed.

Lemma upd_lookup : forall (m m' : map) (a : A) (b : B),
  upd m a b = SomeX m'
  -> lookup m' a b.
Proof.
  intros m m' a b H.
  apply get_lookup.
  assert (H0 := upd_sem m a b).
  destruct (get_dec m a).
  rewrite H in H0. 
  injection H0; intro H1. 
  rewrite H1. 
  apply set_a_get_a.
  rewrite H in H0; discriminate.
Qed.

Lemma canupd_lookup : forall (m m' : map) (a : A) (b : B),
  canupd m a b m' -> lookup m' a b.
Proof.
  intros; unfold canupd in H. eapply upd_lookup; eauto.
Qed.

Lemma canupd_lookup_ex : forall (m m' : map) (a : A) (b : B),
  canupd m a b m' -> exists b', lookup m a b'.
Proof.
  intros.
  assert(H0 := canupd_get H).
  destruct H0 as [b' Hget].
  exists b'; apply get_lookup; trivial.
Qed.

Lemma lookup_canupd : forall (m : map) (a : A) (b b': B),
  lookup m a b -> canupd m a b' (set m a b').
Proof.
  intros.
  unfold canupd.
  eapply get_a_upd_a.
  apply lookup_get.
  eauto.
Qed.

Lemma canupd_lookup_neq : 
  forall (m m' : map) (a a' : A) (b b' : B),
    a <> a' 
    -> canupd m a' b' m' 
    -> lookup m a b 
    -> lookup m' a b.
Proof.
  unfold canupd, lookup; intros.
  assert (H2 : a' <> a).
  intro Hf; subst a; apply H; trivial.
  substH H0 with (upd_a_get_a' H2 H0).
  rewrite H0; trivial.
Qed.

End MapEx.

Unset Implicit Arguments.
