(***************************************************************)
(* Coq code for OCAP                                           *)
(*                                                             *)
(* Utilities for manipulating natural numbers                  *)
(*                                                             *)
(*                                                             *)
(***************************************************************)

(* $Id: bnat.v,v 1.4 2006/07/15 02:56:42 ssg05 Exp $ *)

Require Export EqNat.
Require Export Lt.
Require Export Le.
Require Export Bool.
Require Import Omega.

Fixpoint blt_nat (n m : nat) {struct n} : bool :=
  match n, m with
  | O, O => false
  | O, S _ => true
  | S _, O => false
  | S n', S m' => blt_nat n' m'
  end.

Lemma blt_irrefl :
  forall a : nat, blt_nat a a = false.
Proof.
  induction a; simpl; auto with arith.
Qed.

Lemma blt_irrefl_Prop :
  forall a : nat, ~ (blt_nat a a = true).
Proof.
  induction a; simpl; auto with arith.
Qed.

Lemma blt_asym :
  forall a b : nat, blt_nat a b = true
    -> blt_nat b a = false.
Proof.
  double induction a b; simpl; intros; auto with arith.
Qed.

Lemma blt_asym_Prop :
  forall a b : nat, blt_nat a b = true 
    -> blt_nat b a = true
    -> False.
Proof.
  double induction a b; simpl; intros; auto with arith;
    try discriminate.
  apply H0 with n; trivial.
Qed.

Lemma blt_0_Sn : 
  forall n : nat, blt_nat 0 (S n) = true.
Proof.
  induction n; simpl; auto with arith.
Qed.

Lemma not_blt_n_O :
  forall n : nat, ~ (blt_nat n 0 = true).
Proof.
  induction n; simpl; auto with arith.
Qed.

Lemma blt_lt : 
  forall a b : nat, blt_nat a b = true -> a < b.
Proof.
  double induction a b; simpl; intros; 
    auto with arith; discriminate.
Qed.

Lemma blt_f_le :
  forall n m, blt_nat n m = false -> m <= n.
Proof.
  double induction n m; simpl; intros; 
    auto with arith; discriminate.
Qed.

Lemma le_blt_false :
  forall n m, n <= m -> blt_nat m n = false.
Proof.
  double induction n m; simpl; intros; 
    auto with arith.
  destruct (lt_n_O _ H0).
Qed.

Lemma lt_blt : 
  forall a b : nat, a < b -> blt_nat a b = true.
Proof.
  double induction a b; simpl; intros; 
    auto with arith.
  destruct (lt_irrefl 0 H).
  destruct (lt_n_O (S n) H0).
Qed.

Lemma blt_n_Sn : 
  forall n : nat, blt_nat n (S n) = true.
Proof.
  induction n; simpl; intros; auto with arith.
Qed.

Lemma blt_S_eq :
  forall a b, blt_nat a b = blt_nat (S a) (S b).
Proof.
  trivial.
Qed.

Lemma blt_n_Sm :
  forall n m, blt_nat n m = true 
    -> blt_nat n (S m) = true.
Proof.
  double induction n m; simpl; intros; auto with arith.
  discriminate.
Qed.

Lemma blt_n_mk :
  forall n m k, blt_nat n m = true
    -> blt_nat n (m + k) = true.
Proof.
  double induction n m; simpl; intros; 
    auto with arith;  discriminate.
Qed.

Lemma blt_n_km :
  forall n m k, blt_nat n m = true
    -> blt_nat n (k + m) = true.
Proof.
  double induction n m; simpl; intros; 
    auto with arith; try discriminate.
  replace (k + (S n0)) with ((S n0) + k);
    [idtac | omega].
  simpl. trivial.
  replace (k + (S n0)) with ((S n0) + k);
    [idtac | omega].
  simpl. 
  replace (n0 + k) with (k + n0);
    [idtac | omega].
  apply H0; trivial.
Qed.

Lemma blt_t_bgt_f : 
  forall n m, blt_nat n m = true 
    -> blt_nat m n = false.
Proof.
  double induction n m; simpl; intros; auto with arith.
Qed.

Lemma blt_t_beq_f : 
  forall n m, blt_nat n m = true 
    -> beq_nat m n = false.
Proof.
  double induction n m; simpl; intros; auto with arith.
Qed.

Lemma blt_beq_t_blt_f :
  forall n m, blt_nat n m = true \/ beq_nat n m = true
    -> blt_nat m n = false.
Proof.
  double induction n m; simpl; intros; auto with arith.
  destruct H0; discriminate.
Qed.

Lemma blt_f_blt_beq_t : 
  forall n m, blt_nat n m = false 
    -> blt_nat m n = true \/ beq_nat m n = true.
Proof.
  double induction n m; simpl; intros; auto with arith.
Qed.

Lemma bgt_t_beq_f :
  forall n m, blt_nat m n = true
    -> beq_nat n m = false.
Proof.
  double induction n m; simpl; intros; auto with arith.
Qed.

Lemma beq_eq :
  forall n m, beq_nat n m = true
    -> n = m.
Proof. 
  double induction n m; simpl; intros; 
    auto with arith; discriminate.
Qed.

Lemma beq_false_neq :
  forall n m, beq_nat n m = false
    -> ~ (n = m).
Proof.
  double induction n m; simpl; intros; 
    auto with arith; discriminate.
Qed.

Lemma eq_beq :
  forall n m, n = m
    -> beq_nat n m = true.
Proof. 
  double induction n m; simpl; intros; 
    auto with arith; discriminate.
Qed.

Lemma neq_beq_false :
  forall n m, ~ (n = m) 
    -> beq_nat n m = false.
Proof.
  double induction n m; simpl; intros; 
    auto with arith.
  destruct (H (refl_equal _)).
Qed.

Lemma blt_beq_bgt_dec : forall a b,
  {blt_nat a b = true /\ beq_nat a b = false /\ blt_nat b a = false}
  + {blt_nat a b = false /\ beq_nat a b = true /\ blt_nat b a = false}
  + {blt_nat a b = false /\ beq_nat a b = false /\ blt_nat b a = true}.
Proof.
  double induction a b; simpl; intros; try tauto || auto.
Qed.

Lemma blt_nat_dec : forall a b,
  {blt_nat a b = true} + {blt_nat a b = false}.
Proof.
  double induction a b; simpl; intros; try tauto || auto.
Qed.

Lemma blt_bge_dec : forall a b,
  {blt_nat a b = true} + {blt_nat a b = false}.
Proof.
  double induction a b; simpl; intros; try tauto || auto.
Qed.

Lemma beq_dec : forall a b,
  {beq_nat a b = true} + {beq_nat a b = false}.
Proof.
  double induction a b; simpl; intros; try tauto || auto.
Qed.

Ltac bnat_dec k n := 
  let Hb := fresh "Hb" in
  let Hblt := fresh "Hblt" in
  let Hbeq := fresh "Hbeq" in
  let Hbgt := fresh "Hbgt" in
    (destruct (blt_beq_bgt_dec k n) as [[Hb | Hb] | Hb];
      destruct Hb as [Hblt [Hbeq Hbgt]]).

Ltac bnat_elim_auto :=
  match goal with
    | H : blt_nat ?a ?b = ?c 
      |- context [blt_nat ?a ?b] =>
        rewrite H

    | H : blt_nat ?a ?b = ?c,
      H0 : context [blt_nat ?a ?b]
      |- _ =>
        rewrite H in H0

    | H : beq_nat ?a ?b = ?c
      |- context [beq_nat ?a ?b] =>
        rewrite H

    | H : beq_nat ?a ?b = true |- _ =>
        let H := fresh "H" in
          (generalize (beq_eq a b H); intro H;
            rewrite H)
  end.

(* from cxlin's code *)
Ltac simplbeq := 
  match goal with
    | [ |- context [(beq_nat ?x ?x)] ] => 
      rewrite (eq_beq x x (refl_equal x))
    | [ H : ?x <> ?y |- context [(beq_nat ?x ?y)] ] => 
      rewrite (neq_beq_false x y H)
    | [ H : ?y <> ?x |- context [(beq_nat ?x ?y)] ] => 
      rewrite (neq_beq_false x y (sym_not_eq H))
    | [ |- context [(beq_nat ?x ?y)] ] => 
      rewrite neq_beq_false
  end.

(* Ltac des_beq a b := *)
(*   match goal with *)
(*     | _ |- context *)
(*   end. *)

Ltac bnat_blt_dec k n := 
  let Hb := fresh "Hb" in
  let Hblt := fresh "Hblt" in
  let Hnblt := fresh "Hnblt" in
    (destruct (blt_bge_dec k n) as [Hblt | Hnblt]; 
    repeat bnat_elim_auto).


Ltac bnat_rewrite_hyps :=
  match goal with
    | H : blt_nat ?a ?b = ?c 
      |- context [blt_nat ?a ?b] =>
        rewrite H

    | H : beq_nat ?a ?b = ?c
      |- context [beq_nat ?a ?b] =>
        rewrite H

    | H : beq_nat ?a ?b = true |- _ =>
        let H := fresh "H" in
          (generalize (beq_eq a b H); intro H;
            rewrite H)
  end.


Ltac bnat_elim := 
  match goal with
    | H : blt_nat ?a ?b = true |- _ =>
        generalize (blt_lt a b H); clear H; intro H

    | H : blt_nat ?a ?b = false |- _ =>
        generalize (blt_f_le a b H); clear H; intro H

    | H : beq_nat ?a ?b = true |- _ =>
        generalize (beq_eq a b H); clear H; intro H

    | H : beq_nat ?a ?b = false |- _ =>
        generalize (beq_false_neq a b H); clear H; intro H

    | |- blt_nat ?a ?b = true =>
        apply (lt_blt a b)

    | |- blt_nat ?a ?b = false =>
        apply (le_blt_false b a)

    | |- beq_nat ?a ?b = true =>
        apply (eq_beq a b)

    | |- beq_nat ?a ?b = false =>
        apply (neq_beq_false a b)

    | _ => fail 1 "no ble_nat hyp found"
  end.

Ltac bnat_rev :=
  repeat bnat_elim; omega.

Ltac bnat_discr :=
  match goal with 

    | H: blt_nat ?a ?b = true, 
      H1 : blt_nat ?a ?b = false 
      |- _ => 
        rewrite H in H1; discriminate

    | H: blt_nat ?a ?b = true, 
      H1 : blt_nat ?b ?a = true
      |- _ => 
        destruct (blt_asym_Prop a b H H1)

    | H : blt_nat ?a ?a = true 
      |- _ =>
        destruct (blt_irrefl_Prop a H)

    | H : blt_nat ?a O = true 
      |- _ =>
        destruct (not_blt_n_O a H)
        
    | H: beq_nat ?a ?b = true, 
      H1 : beq_nat ?a ?b = false 
      |- _ => 
        rewrite H in H1; discriminate

    | _ => fail 1 "no discriminatable hyptheses"
  end.

Tactic Notation "rep_arith" constr (t1) "with" constr (t2) := 
  replace t1 with t2; [idtac | repeat bnat_elim; omega].

Tactic Notation "rep_arith" constr (t1) "with" constr (t2) "in" hyp (H) := 
  replace t1 with t2 in H; [idtac | repeat bnat_elim; omega].

Tactic Notation "bnat_rew" := bnat_rewrite_hyps.

Tactic Notation "bnat_rew" constr (t) :=
  let H := fresh "H" in
    (assert (H : t); [bnat_rev | rewrite H; clear H]).

Tactic Notation "bnat_rew" constr (t) "in" hyp (H0) :=
  let H := fresh "H" in
    (assert (H : t); [bnat_rev | rewrite H in H0; clear H]).


(*
Ltac ble_le :=
  match goal with
    | H : ble_nat ?a ?b = true |- _ =>
        generalize (ble_true_le a b H); clear H; intro H
    | H : ble_nat ?a ?b = false |- _ =>
        generalize (ble_false_le a b H); clear H; intro H
    | H : ?a <= ?b |- context [ble_nat ?a ?b] =>
        rewrite (le_ble_true a b H); cbv iota
    | H : ?a <= ?b |- context [ble_nat ?b ?a] =>
        rewrite (le_ble_true b a H); cbv iota
    | H : ?a <= ?b, H0 : context [ble_nat ?a ?b] |- _ =>
        rewrite (le_ble_true a b H) in H0; cbv iota in H0
    | H : ?a <= ?b, H0 : context [ble_nat ?b ?a] |- _ =>
        rewrite (le_ble_true b a H) in H0; cbv iota in H0
    | |- ble_nat ?a ?b = true =>
        apply (le_ble_true a b)
    | |- ble_nat ?a ?b = false =>
        apply (le_ble_false a b)
    | _ => fail 1 "no ble_nat hyp found"
  end.
*)
